"""
Strategist service. --- :mod:`alchemiscale.strategist.service`
==============================================================


Applies `Strategy`\s to `AlchemicalNetwork`\s to populate `TaskHubs`\s on .

"""
